;; Soundscrape
;; Copyright (C) 2003,2004 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or    
;; modify it under the terms of the GNU General Public License as   
;; published by the Free Software Foundation; either version 2 of   
;; the License, or (at your option) any later version.              
;;                                                                  
;; This program is distributed in the hope that it will be useful,  
;; but WITHOUT ANY WARRANTY; without even the implied warranty of   
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    
;; GNU General Public License for more details.                     
;;                                                                  
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, contact:
;;
;; Free Software Foundation           Voice:  +1-617-542-5942
;; 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
;; Boston, MA  02111-1307,  USA       gnu@gnu.org

;;; Commentary:
;;
;;Blah?
;;
;;; Code:

(define-module (soundscrape ui pages)
  :use-module (gnome gtk)
  :use-module ((srfi srfi-1) :select (fold lset-difference))
  :use-module (soundscrape util)
  :use-module (soundscrape core)
  :use-module (soundscrape ui util)
  :use-module (gnome glade)

  :export (define-page-type do-action on-switch-page set-actions-enabled!
           initialize-action-widgets action-helper add-page
           set-page-title! page-name page-widget page-type remove-page!))

;; Very tricky macro!
(define-macro (define-page-type name . actions)
  `(define ,name
     (cons '%page-type%
           ,(let lp ((in actions) (out `((stop . ,on-stop))))
              (cond ((null? in) (list 'quasiquote out))
                    ((null? (cdr in)) (error "invalid actions"))
                    (else
                     (lp (cddr in)
                         (cons (cons (car in) (list 'unquote (cadr in)))
                               out))))))))

(define action-widgets #f)

(define (set-actions-enabled! . args) ; action1 enabled1? action2 enabled2? ...
  (let lp ((args args))
    (unless (null? args)
            (begin
              (for-each
               (lambda (x) (set x 'sensitive (cadr args)))
               (or (assq-ref action-widgets (car args))
                   (error "Mayday, mayday!" (car args) action-widgets)))
              (lp (cddr args))))))

(define (initialize-action-widgets widget action-list)
  (let ((xml (glade-get-widget-tree widget)))
    (set! action-widgets
          (map (lambda (action)
                 (cons (car action)
                       (map (lambda (s) (get-widget xml s)) (cdr action))))
               action-list))))

(define the-notebook #f)
(define all-pages '()) ; index in this list may change.

(define current-page #f) ; (<widget> "tab name" type)

(define (do-action name . args)
  (apply (or (assq-ref (cdr (page-type current-page)) name)
             (error "Action not supported" name current-page))
         current-page
         args))

(define (action-helper action)
  (lambda args
    (apply do-action action args)))

(define (make-page widget name page-type)
  (list widget name page-type))
(define (page-widget page)
  (car page))
(define (page-name page)
  (cadr page))
(define (page-type page)
  (caddr page))
(define (page? page) ;; very loose predicate
  (and (list? page) (eq? (length page) 3)))

(define (list-index* l k eq?)
  (let loop ((l l) (n 0))
    (cond ((null? l) #f)
          ((eq? (car l) k) n)
          (else (loop (cdr l) (1+ n))))))

(define (->page x)
  (cond
   ((string? x)
    (list-ref all-pages (list-index (map page-name all-pages) x)))
   ((is-a? x <gtk-widget>)
    (list-ref all-pages (list-index* (map page-widget all-pages) x equal?)))
   ((page? x)
    x)
   (else
    (error "Don't know how to find page from argument" x))))

(define (on-stop page . args)
  (ss-engine-stop!))

(add-hook! ss-engine-run-hook
           (lambda (running?)
             (set-actions-enabled! 'stop running?)))

(define (switch-to-page page)
  (let* ((page (->page page))
         (n (page-num the-notebook (page-widget page))))
    (set-current-page the-notebook n)))

(define (on-switch-page notebook child n)
  ;; the switch-page signal is broken: `child' is declared as a
  ;; GPointer, not a GtkWidget. Therefore only rely on `n'.
  (let ((page (->page (get-nth-page notebook n))))
    (set! current-page page)
    ;; Can't think of an efficient way to do this without changing from
    ;; alists to hash tables or something.
    (do-action 'switch-to-page)
    (set-actions-enabled! 'stop (ss-engine-running?))
    (apply set-actions-enabled!
           (fold (lambda (x val) (cons* x #f val)) '()
                 (lset-difference eq?
                                  (map car action-widgets)
                                  (map car (cdr (page-type current-page))))))))

(define-macro (with-current-page page . body)
  (let ((old-page (gensym)))
    `(let ((,old-page current-page))
       (dynamic-wind
           (lambda () (set! current-page ,page))
           (lambda () ,@body)
           (lambda () (set! current-page ,old-page))))))

(define (add-page type name widget)
  (unless the-notebook
          (set! the-notebook 
                (get-widget (glade-get-widget-tree 
                             (slot-ref the-main-window 'gtk-window))
                            "tabs")))
  (let* ((box (make <gtk-hbox> :spacing 6))
         (text (make <gtk-label> :label name))
         (page (make-page widget name type)))
    (pack-start box text)

    (if (assq 'close (cdr type))
        (let ((image (make <gtk-image> :stock "gtk-close"
                           :icon-size (genum->value
                                       (make <gtk-icon-size> :value 'menu))))
              (button (make <gtk-button>)))
          (pack-start box button)
          (add button image)
          (connect button 'clicked
                   (lambda args
                     (with-current-page page (do-action 'close))))))

    (show-all box)
    (set! all-pages (cons page all-pages))
    (append-page the-notebook widget box)
    (switch-to-page page)
    page))

(define (remove-page! widget)
  (let ((page (->page widget)))
    (remove-page the-notebook (page-num the-notebook widget))
    (set! all-pages (delq page all-pages))))

(define (set-page-title! x title)
  (let ((page (->page x)))
    (list-set! page 1 title)
    (set (car (get-children (get-tab-label the-notebook (page-widget page))))
         'label title)))
